"""
A system-wide Lock class.
Where the system is a server, computer, container, and/or file system.
"""
from fcntl import LOCK_EX, LOCK_NB, LOCK_UN, flock

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING
from infrahouse_toolkit.lock.base import BaseLock
from infrahouse_toolkit.lock.exceptions import LockAcquireError


class SystemLock(BaseLock):
    """
    Acquire a system-wide lock.

    :param lockfile: Path to a file on the local filesystem.
    :type lockfile: str
    :param blocking: Whether the caller should wait for the lock or exception
        should be raised. If blocking is True, the caller will wait.
    :type blocking: bool
    :raise LockAcquireError: If we tried to acquire the lock in a non-blocking mode
        and the lock is taken by somebody else.
    """

    def __init__(self, lockfile: str, blocking: bool = True):
        self._lockfile = lockfile
        self._blocking = blocking
        self._f_desc = None
        super().__init__()
        self._name = f"{self.__class__.name}@{self._lockfile}"

    def __enter__(self):
        self._f_desc = open(self._lockfile, "a", encoding=DEFAULT_OPEN_ENCODING)
        try:
            flock(self._f_desc, LOCK_EX if self._blocking else LOCK_EX | LOCK_NB)
        except BlockingIOError as exc:
            raise LockAcquireError(f"Failed to get an exclusive lock on {self._lockfile}") from exc

    def __exit__(self, exc_type, exc_val, exc_tb):
        flock(self._f_desc, LOCK_UN)



"""
.. topic:: ``ih-s3-reprepro includedeb``

    A ``ih-s3-reprepro includedeb`` subcommand.

    See ``ih-s3-reprepro includedeb --help`` for more details.
"""
import sys

import click
from botocore.exceptions import BotoCoreError, ClientError
from click import Context

from infrahouse_toolkit import LOG
from infrahouse_toolkit.cli.ih_s3_reprepro.utils import execute, repo_env


@click.command(name="includedeb")
@click.argument("codename")
@click.argument("deb_file")
@click.pass_context
def cmd_includedeb(ctx: Context, codename, deb_file):
    """
    Include the given binary package.

    Include the given binary Debian package (.deb) in the specified distribution,
    applying override information and guessing all values not given and guessable.
    """
    try:
        with repo_env(
            ctx.parent.params["bucket"],
            ctx.parent.params["role_arn"],
            ctx.parent.params["gpg_key_secret_id"],
            ctx.parent.params["gpg_passphrase_secret_id"],
            region=ctx.parent.params["aws_region"],
        ) as (path, gpg_home):
            execute(["reprepro", "-V", "-b", path, "--gnupghome", gpg_home, "includedeb", codename, deb_file])

    except (ClientError, BotoCoreError) as err:
        LOG.error(err)
        sys.exit(1)



import pytest

from infrahouse_toolkit.lock.exceptions import LockAcquireError
from infrahouse_toolkit.lock.system import SystemLock
from infrahouse_toolkit.timeout import timeout


def test_system_lock(tmpdir):
    lock_file = tmpdir.join("lock")
    with pytest.raises(TimeoutError):
        with timeout(1):
            with SystemLock(str(lock_file)):
                with SystemLock(str(lock_file)):
                    pass


@pytest.mark.timeout(3)
def test_system_lock_non_blocking(tmpdir):
    lock_file = tmpdir.join("lock")
    with pytest.raises(LockAcquireError):
        with SystemLock(str(lock_file)):
            with SystemLock(str(lock_file), blocking=False):
                pass



from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.status import RunOutput, RunResult, TFStatus


def test_eq():
    assert TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(1, 1, 1), RunOutput("no stdout", None)
    ) == TFStatus(TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(1, 1, 1), RunOutput("no stdout", None))


def test_eq_no_affrected_res():
    assert TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"),
        True,
        RunResult(1, 1, 1),
        RunOutput("no stdout", None),
        affected_resources=RunResult(None, None, None),
    ) == TFStatus(TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(1, 1, 1), RunOutput("no stdout", None))


def test_neq():
    assert TFStatus(
        TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(1, 1, 1), RunOutput("no stdout", None)
    ) != TFStatus(TFS3Backend("foo_backet", "path/to/tf.state"), True, RunResult(1, 0, 1), RunOutput("no stdout", None))



"""
.. topic:: ``ih-s3-reprepro set-secret-value``

    A ``ih-s3-reprepro set-secret-value`` subcommand.

    See ``ih-s3-reprepro set-secret-value --help`` for more details.
"""
import sys

import click
from botocore.exceptions import BotoCoreError, ClientError
from click import Context

from infrahouse_toolkit import DEFAULT_OPEN_ENCODING, LOG
from infrahouse_toolkit.cli.ih_s3_reprepro.aws import get_client


@click.command(name="set-secret-value")
@click.argument("secret_id")
@click.argument("path", nargs=-1)
@click.pass_context
def cmd_set_secret_value(ctx: Context, secret_id, path):
    """
    Set value to a secret.

    Optionally the value may be given via a local file specified by a path argument.

    \b
    ih-s3-reprepro ... set-secret-value mysecret /path/to/file_with_value

    if the path is omitted, a user will be prompt for the value.
    """
    role_arn = ctx.parent.params["role_arn"]
    try:
        client = get_client("secretsmanager", role_arn=role_arn, region=ctx.parent.params["aws_region"])
        if path:
            with open(path[0], encoding=DEFAULT_OPEN_ENCODING) as secret_value_desc:
                value = secret_value_desc.read()
        else:
            value = click.prompt("Enter a secret value and press ENTER", hide_input=True)

        LOG.debug("Secret value: %s", value)
        client.put_secret_value(SecretId=secret_id, SecretString=value)
        LOG.info("Value of %s is successfully set.", secret_id)

    except (ClientError, BotoCoreError) as err:
        LOG.error(err)
        sys.exit(1)



from infrahouse_toolkit.cli.ih_plan.cmd_min_permissions import ActionList


def test_add_sorted():
    """Ensure sorted order"""
    actions1 = ActionList()
    actions1.add("foo")
    actions1.add("bar")

    actions2 = ActionList()
    actions2.add("bar")
    actions2.add("foo")

    assert actions1.actions == actions2.actions


def test_add():
    """add() an action and maintain uniqueness."""
    actions = ActionList()
    actions.add("foo")
    assert actions.actions == ["foo"]

    actions.add("foo")
    assert actions.actions == ["foo"]

    actions.add("bar")
    assert "bar" in actions.actions
    assert "foo" in actions.actions


def test_add_with_dependency():
    actions = ActionList()
    actions.add("autoscaling:CreateAutoScalingGroup")

    assert "autoscaling:CreateAutoScalingGroup" in actions.actions
    assert "iam:PassRole" in actions.actions
    assert "iam:CreateServiceLinkedRole" in actions.actions


def test_add_with_rewrite():
    actions = ActionList()
    actions.add("auto scaling:foo")

    assert actions.actions == ["autoscaling:foo"]



"""
Module for :py:class:`TFBackend`.
A generic class that defines the API for all kind of Terraform backends.
"""

from abc import ABC, abstractmethod


class TFBackend(ABC):
    """API to a Terraform backend."""

    # pylint: disable=invalid-name,too-few-public-methods
    # this is an abstract class, so it's OK to have fewer methods/attributes.
    @property
    @abstractmethod
    def id(self) -> str:  # pylint: disable=invalid-name
        """
        A unique identifier of the Terraform backend.
        """



"""
.. topic:: ``ih-puppet apply``

    A ``ih-puppet apply`` subcommand.

    See ``ih-puppet apply`` for more details.
"""
import sys
from os import environ
from subprocess import Popen

import click

from infrahouse_toolkit import LOG
from infrahouse_toolkit.lock.system import SystemLock


@click.command(name="apply")
@click.argument("manifest", required=False)
@click.pass_context
def cmd_apply(ctx, manifest):
    """
    Apply puppet manifest
    """
    manifest = manifest or f"{ctx.obj['root_directory']}/environments/{ctx.obj['environment']}/manifests/site.pp"
    LOG.info("Applying puppet manifest in %s", manifest)
    cmd = ["puppet", "apply"]
    if ctx.obj["debug"]:
        cmd.append("-d")

    cmd.extend(
        [
            "--environment",
            ctx.obj["environment"],
            "--hiera_config",
            ctx.obj["hiera_config"],
            "--hiera_config",
            ctx.obj["hiera_config"],
            f"--modulepath={ctx.obj['module_path']}",
            manifest,
            "--write-catalog-summary",
            "--detailed-exitcodes",
        ]
    )

    with SystemLock("/var/run/ih-puppet-apply.lock"):
        LOG.debug("Executing %s", " ".join(cmd))
        env = {"PATH": f"{environ['PATH']}:/opt/puppetlabs/bin"}
        # First run is to update the puppet code
        with Popen(cmd, env=env) as proc:
            proc.communicate()

        # Second run is to apply whatever the new puppet code brings
        with Popen(cmd, env=env) as proc:
            proc.communicate()
            ret = proc.returncode
            LOG.debug("Exit code: %d", ret)
            if ret == 0:
                LOG.info("The run succeeded with no changes or failures; the system was already in the desired state.")
                sys.exit(0)
            elif ret == 1:
                LOG.error("The run failed.")
                sys.exit(ret)
            elif ret == 2:
                LOG.info("The run succeeded, and some resources were changed.")
                sys.exit(0)
            elif ret == 4:
                LOG.warning("The run succeeded, and some resources failed.")
                sys.exit(ret)
            elif ret == 6:
                LOG.warning("The run succeeded, and included both changes and failures.")
                sys.exit(ret)
            else:
                LOG.error("Unknown run state.")
                sys.exit(ret)



from os import environ
from unittest import mock
from unittest.mock import Mock, call

import pytest
from github import GithubException

from infrahouse_toolkit.terraform import RunOutput, TFStatus, parse_plan
from infrahouse_toolkit.terraform.backends import TFS3Backend
from infrahouse_toolkit.terraform.githubpr import GitHubPR


@pytest.mark.skipif("GITHUB_TOKEN" not in environ, reason="This is a development test, needs GITHUB_TOKEN")
def test_publish_comment():
    gh_pr = GitHubPR("infrahouse8/github-control", 33)
    gh_pr.delete_my_comments()
    counts, resources = parse_plan(open("infrahouse_toolkit/terraform/tests/plans/plan-2-1-2.stdout").read())
    status = TFStatus(
        TFS3Backend("foo-bucket", "path/to/key.state"),
        success=True,
        run_result=counts,
        run_output=RunOutput(open("infrahouse_toolkit/terraform/tests/plans/plan-2-1-2.stdout").read(), None),
        affected_resources=resources,
    )
    print(status.comment)
    gh_pr.publish_comment(status.comment)


def test_publish_too_large_comment():
    gh_pr = GitHubPR("foo/bar", 123)
    mock_pull_request = Mock()
    mock_pull_request.create_issue_comment.side_effect = [GithubException(422, "some data", {}), None]

    mock_gist = Mock()
    mock_gist.html_url = "gist url"
    with mock.patch.object(
        GitHubPR, "pull_request", new_callable=mock.PropertyMock, return_value=mock_pull_request
    ), mock.patch.object(GitHubPR, "_publish_gist", return_value=mock_gist) as mock_publish_gist:
        gh_pr.publish_comment("foo comment")
        mock_publish_gist.assert_called_once_with("pr-123-plan", "foo-bar-pr-123-plan.txt", "foo comment")
        mock_pull_request.create_issue_comment.assert_has_calls(
            [call("foo comment"), call("Comment was too big. It's published as a gist at gist url.")]
        )



"""
:py:mod:`infrahouse_toolkit.terraform` exceptions.
"""
from infrahouse_toolkit.exceptions import IHException


class IHTFException(IHException):
    """Terraform related exceptions."""


class IHParseError(IHTFException):
    """Error happening when parsing fails."""



"""Unit tests for :py:func:`infrahouse_toolkit.cli.lib.get_bucket`."""
from infrahouse_toolkit.cli.lib import get_bucket


def test_get_bucket(terraform_tf):
    """Check that :py:func:`infrahouse_toolkit.cli.lib.get_bucket` can parse a valid ``terraform.tf``."""
    assert get_bucket(tf_file=terraform_tf) == "infrahouse-foo"



from infrahouse_toolkit.terraform.githubpr import GitHubPR


def test_init():
    gh_pr = GitHubPR("infrahouse8/github-control", 33)
    for comment in gh_pr.comments:
        # 1/0
        print(comment.body)
